/***************************************************************************
 *   Copyright (C) 2007 by Veselin Georgiev                                *
 *   anrieff@mgail.com (convert to GMail)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "utils.h"
#include "bitmap.h"
#include "embedded_bitmaps.h"

#if __BYTE_ORDER == __LITTLE_ENDIAN
#	define itohl(x) (x)
#	define itohs(x) (x)
#	define htoil(x) (x)
#	define htois(x) (x)
#else
#	define itohl(x) byteswap_32(x)
#	define itohs(x) byteswap_16(x)
#	define htoil(x) byteswap_32(x)
#	define htois(x) byteswap_16(x)
#endif


uint32_t byteswap_32(uint32_t x)
{
	x = ((x & 0xffff) << 16) | ((x >> 16) & 0xffff);
	x = ((x & 0x00ff00ff) << 8) | ((x >> 8) & 0x00ff00ff);
	return x;
}

uint16_t byteswap_16(uint16_t x)
{
	x = ((x & 0xff) << 8) | ((x >> 8) & 0xff);
	return x;
}


// the 'BM' magic in the beginning of a .BMP phile
#define BM_MAGIC 19778

// the .BMP header structure is taken from http://www.fortunecity.com/skyscraper/windows/364/bmpffrmt.html
typedef struct {
	int fs;       // filesize
	int lzero;
	int bfImgOffset;  // basic header size
}BMHEADER;

typedef struct {
	int ihdrsize; 	// info header size
	int x,y;      	// image dimensions
	uint16_t channels;// number of planes
	uint16_t bitsperpixel;
	int compression; // 0 = no compression
	int biSizeImage; // used for compression; otherwise 0
	int pixPerMeterX, pixPerMeterY; // dots per meter settings (why meters and not inches... maybe an attempt to conform to SI? :)
	int colors;	 // number of used colors. If all specified by the bitsize are used, then it should be 0
	int colorsImportant; // number of "important" colors (All in wonder..)
}
#ifdef __GNUC__
__attribute__((packed))
#endif
BMiHEADER;


class InputStream {
public:
	virtual int read(void * buff, int howmuch) = 0;
	virtual ~InputStream() {}
};

class FileInputStream: public InputStream {
	FILE *f;
public:
	FileInputStream() { f = NULL; }
	bool open_file(const char * filename)
	{
		f = fopen(filename, "rb");
		return f != NULL;
	}
	int read(void *buff, int howmuch)
	{
		if (!f) return 0;
		int acread = (int) fread(buff, (size_t) 1, (size_t)howmuch, f);
		return acread;
	}
	~FileInputStream()
	{
		if (f)
			fclose(f);
		f = NULL;
	}
};

class MemoryInputStream: public InputStream {
	unsigned char * data;
	int ptr, size;
public:
	MemoryInputStream(unsigned char *dta, int sz)
	{
		data = dta;
		size = sz;
		ptr = 0;
	}
	int read(void *buff, int howmuch)
	{
		int i = 0;
		unsigned char *p = (unsigned char*) buff; 
		while (ptr < size && i < howmuch) {
			p[i] = data[ptr];
			i++, ptr++;
		}
		return i;
	}
	~MemoryInputStream(){}
};

Bitmap::Bitmap()
{
	x = y = 0;
	d = NULL;
}

Bitmap::~Bitmap()
{
	clear();
}

Bitmap::Bitmap(int x, int y)
{
	this->x = x;
	this->y = y;
	d = new unsigned[x*y];
	memset(d, 0, sizeof(unsigned) * x * y);
}

void Bitmap::clear(void)
{
	if (d) delete [] d;
	d = NULL;
	x = y = 0;
}

int Bitmap::width() const { return x;}
int Bitmap::height() const { return y;}
unsigned * Bitmap::data() const { return d; }

bool Bitmap::load_from_file(const char *filename)
{
	FileInputStream fi;
	if (!fi.open_file(filename)) return false;
	return _loadimp(&fi);
}

bool Bitmap::load_embedded(const char *name)
{
	int n = (int) (sizeof(embedded_bitmaps) / sizeof(embedded_bitmaps[0]));
	for (int i = 0; i < n; i++)
		if (!strcmp(embedded_bitmaps[i].name, name)) {
			MemoryInputStream mi((unsigned char*)embedded_bitmaps[i].data, embedded_bitmaps[i].size);
			return _loadimp(&mi);
		}
	return false;
}

bool Bitmap::save_to_file(const char *filename)
{
	FILE *f;
	int i, j, rowsz;
	unsigned short msh=BM_MAGIC;
	BMHEADER hd;
	BMiHEADER hi;
	char xx[2222*3];
	
	
	if ((f=fopen(filename, "wb"))==NULL) {
		printf("SaveBmp:can't open `%s' for writing, check permissions!\n", filename);
		return false;
	}
	// fill in the header:
	rowsz = x * 3;
	if (rowsz % 4)
		rowsz += 4 - (rowsz % 4); // each row in of the image should be filled with zeroes to the next multiple-of-four boundary
	hd.fs = rowsz*y + 54; //std image size :)
	hd.lzero = 0;
	hd.bfImgOffset = 54;
	hi.ihdrsize = 40;
	hi.x = x; hi.y = y;
	hi.channels = 1;
	hi.bitsperpixel = 24; //RGB format
	// set the rest of the header to default values:
	hi.compression = hi.biSizeImage = 0;
	hi.pixPerMeterX = hi.pixPerMeterY = 0;
	hi.colors = hi.colorsImportant = 0;
	
	fwrite(&msh, 2, 1, f); // write 'BM'
	fwrite(&hd, sizeof(hd), 1, f); // write file header
	fwrite(&hi, sizeof(hi), 1, f); // write image header
	int *adata = (int*) d;
	for (j=y-1;j>=0;j--) {
		for (i=0;i<x;i++) {
			xx[i*3  ] = (0xff     & adata[j*x+i]);
			xx[i*3+1] = (0xff00   & adata[j*x+i]) >> 8;
			xx[i*3+2] = (0xff0000 & adata[j*x+i]) >> 16;
		}
		fwrite(xx, rowsz, 1, f);
	}
	fclose(f);
	return true;
}

bool Bitmap::_loadimp(InputStream *f)
{
	int i, j, p, k;
	BMHEADER hd;
	BMiHEADER hi;
	int pal[256];
	int toread = 0;
	unsigned char *xx;
	int rowsz;
	unsigned short sign;
	
	if (!f->read(&sign, 2)) return false;
	if (sign!=itohs(BM_MAGIC)) return false;
	if (!f->read(&hd, sizeof(hd))) return false;
	if (!f->read(&hi, sizeof(hi))) return false;
	
	/* header correctness checks */
	if (!(itohs(hi.bitsperpixel)==8 || itohs(hi.bitsperpixel)==24 || itohs(hi.bitsperpixel)==32)){
		return false;
	}
	if (itohs(hi.channels)!=1) {
		return false;
	}
	/* ****** header is OK *******/
	
	// if image is 8 bits per pixel or less (indexed mode), read some pallete data
	if (itohs(hi.bitsperpixel) <= 8) {
		toread = (1<<itohs(hi.bitsperpixel));
		if (itohl(hi.colors)) toread=itohl(hi.colors);
		if (!f->read(pal, toread* 4)) return false;
	}
	toread = itohl(hd.bfImgOffset) - (54 + toread*4);
	char useless;
	if (toread)
	// skip the rest of the header
		for (i=0;i<toread;i++)
			f->read(&useless, 1);
	k = itohs(hi.bitsperpixel) / 8;
	rowsz = itohl(hi.x) * k;
	if (rowsz%4)
		rowsz = (rowsz / 4 + 1) * 4; // round the row size to the next exact multiple of 4
	xx = (unsigned char *) malloc(rowsz);
	p = itohl(hi.x) * itohl(hi.y);
	x = itohl(hi.x);
	y = itohl(hi.y);
	d = new unsigned[x*y];
	if (!d) {
		free(xx);
		return false;
	}
	int *adata = (int*)d;
	for (j=itohl(hi.y)-1;j>=0;j--) {// bitmaps are saved in inverted y
		if (!f->read(xx, rowsz)) {
			free(xx);
			return false;
		}
		p -= itohl(hi.x);
		for (i=0;i<itohl(hi.x);i++){ // actually read the pixels
			if (itohs(hi.bitsperpixel)>8)
				adata[p++] = xx[i*k] + ((int) xx[i*k+1])*256 + ((int) xx[i*k+2])*65536;
			else
				adata[p++] = pal[xx[i*k]];
		}
		p -= itohl(hi.x);
	}
	free(xx);
	return true;
}

void Bitmap::rescale(int nx, int ny)
{
	if (nx == 0 && ny == 0) return;
	if (nx == 0) {
		nx = x * ny / y;
	}
	if (ny == 0) {
		ny = y * nx / x;
	}
	if (nx == x && ny == y) return;
	unsigned *nd = new unsigned [nx * ny];
	for (int j = 0; j < ny; j++) {
		for (int i = 0; i < nx; i++) {
			nd[j * nx + i] = this->sample_bilinear((float) (i+0.5f) / nx * x - 0.5f, (float) (j+0.5f) / ny * y - 0.5f);
		}
	}
	
	delete [] d;
	d = nd;
	x = nx;
	y = ny;
}

unsigned Bitmap::sample_bilinear(float fx, float fy)
{
	int x0 = (int) floor(fx);
	int y0 = (int) floor(fy);
	fx -= x0;
	fy -= y0;
	unsigned p00 = sample(x0, y0);
	unsigned p01 = sample(x0+1, y0);
	unsigned p10 = sample(x0, y0+1);
	unsigned p11 = sample(x0+1, y0+1);
	return
		mul(p00, (1-fx) * (1-fy)) +
		mul(p01, (  fx) * (1-fy)) +
		mul(p10, (1-fx) * (  fy)) +
		mul(p11, (  fx) * (  fy));
}

unsigned Bitmap::sample(int ix, int iy)
{
	if (!d) return 0;
	if (ix < 0 || iy < 0 || ix >= x || iy >= y) return 0;
	return d[iy * x + ix];
}
