/***************************************************************************
 *   Copyright (C) 2007 by Veselin Georgiev                                *
 *   anrieff@mgail.com (convert to GMail)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "board.h"
#include "ray.h"
#include "utils.h"
#include "config.h"

void get_quadrant(float xi, float yi, float &xo, float &yo, int &quad)
{
	if (yi >= 0) {
		if (xi >= 0) {
			xo = xi;
			yo = yi;
			quad = 0;
		} else {
			xo = yi;
			yo = -xi;
			quad = 1;
		}
	} else {
		if (xi < 0) {
			xo = -xi;
			yo = -yi;
			quad = 2;
		} else {
			xo = -yi;
			yo = xi;
			quad = 3;
		}
	}
}

#undef sqr
static inline double sqr(double x) { return x*x;}
static inline float sqr(float x) { return x*x;}

static inline float fp(float x)
{
	return x - floor(x);
}

static inline int xcomp(float a, float b, float margin)
{
	if (fabs(a-b) < margin) return 0;
	return a < b? -1 : 1;
}

static inline float pith(float x0, float y0, float x1, float y1)
{
	return sqrtf((x1-x0)*(x1-x0) + (y1-y0)*(y1-y0));
}


class Texture {
	public:
		virtual unsigned sample(float x, float y) = 0;
		virtual ~Texture(){}
};

class ProceduralBoardTexture: public Texture {
public:
	unsigned sample(float xi, float yi) {
		#define BSIZE 0.025
		#undef GRID
		#define GRID if (fabs(fp(x)-0.5) < 0.025 || fabs(fp(y)-0.5) < 0.025) return border_color
		#define LGCIRCLE(x, y) if (pith(0,0,x,y)<0.4) return 0xffffff
		#define SMCIRCLE(x0, y0) if (pith(x, y, x0, y0) < 0.32 || pith(y, x, x0, y0) < 0.32) return 0xffffff
		#define BRDCIRCLE(x0, y0) if (pith(x,y,x0,y0) < 0.45) { \
							float d = pith(x,y,x0,y0); \
							return d < 0.4 ? 0xffffff : border_color; }
		float x, y; int quadrant;
		get_quadrant(xi, yi, x, y, quadrant);
		if (x > 7.9 || y > 7.9)
			return 0x3f0000;
		if (x < 1.5-BSIZE && y < 1.5-BSIZE) {
			SMCIRCLE(0.52, 0.0);
			SMCIRCLE(1.18, 0.0);
			SMCIRCLE(1.18, 0.68);
			switch ( xcomp(x, y, 0.025) ) {
				case -1:
					return player_colors[(quadrant+1) % 4];
					case 0: return border_color;
					case 1: return player_colors[quadrant];
				default:
					return 0;
			}
		}
		if (x > 1.5+BSIZE && y > 1.5+BSIZE) {
			if (x < 2.5+BSIZE && y < 2.5+BSIZE) {
				GRID;
				return 0xffffff;
			} else {
				float d = pith(x, y, 4.5, 4.5);
				if (fabs(d-2) < 0.025) return border_color;
				if (d < 2) {
					BRDCIRCLE(3.9, 3.9);
					BRDCIRCLE(5.1, 3.9);
					BRDCIRCLE(3.9, 5.1);
					BRDCIRCLE(5.1, 5.1);
					return player_colors[(quadrant+1)%4];
				}
				return bgnd_color;
			}
		} else {
			if (x > 7.5+BSIZE || y > 7.5+BSIZE) return bgnd_color;
			if (x > 5.7 && x < 7.2 && y > 0.7 && y < 1.3) {
				float xx = fp((x - 5.7) * 2);
				float yy = fabs(y - 1) / 0.3;
				if (xx > yy) return player_colors[quadrant];
			}
			GRID;
			if (y < 0.5 && x < 6.5) {
				LGCIRCLE(fp(x+0.5)-0.5, y);
				return player_colors[quadrant];
			}
			if (x < 0.5 && y < 6.5) {
				LGCIRCLE(x, fp(y+0.5)-0.5);
				return player_colors[(quadrant+1) %4];
			}
			return 0xffffff;
		}
		return bgnd_color;
	}
};

class LookupTexture : public Texture {
	unsigned *data, bmask;
	float xt, yt;
	int size;
public:
	LookupTexture()
	{
		data = NULL;
	}
	~LookupTexture()
	{
		if (data) delete [] data;
		data = NULL;
	}
	void init(Texture & tex, int size)
	{
		this->size = size;
		bmask = size * size - 1;
		xt = size / 16.0f;
		yt = size / 16.0f;
		data = new unsigned[size*size];
		unsigned *p = data;
		float s = 16.0f / size;
		float samples[][2] = {
			{ 0.250f, 0.125f },
			{ 0.125f, 0.500f },
			{ 0.500f, 0.625f },
			{ 0.625f, 0.250f },
			{ 0.375f, 0.375f },
			{ 0.750f, 0.875f },
			{ 0.375f, 0.750f },
			{ 0.500f, 0.375f },
			{ 0.875f, 0.500f },
			{ 0.625f, 0.625f },
		};
		for (int i = 0; i < size; i++) {
			for (int j = 0; j < size; j++, p++) {
				int r = 0, g = 0, b = 0;
				for (unsigned k = 0; k < sizeof(samples) / sizeof(samples[0]); k++) {
					float x = (j + samples[k][0]-0.5f) * s - 8.0f;
					float y = (i + samples[k][1]-0.5f) * s - 8.0f;
					unsigned t = tex.sample(x, y);
					b += t & 0xff; t>>=8;
					g += t & 0xff; t>>=8;
					r += t & 0xff;
				}
				r /= sizeof(samples) / sizeof(samples[0]);
				g /= sizeof(samples) / sizeof(samples[0]);
				b /= sizeof(samples) / sizeof(samples[0]);
				*p = (b) | (g << 8) | (r << 16);
			}
		}
	}
	unsigned sample(float x, float y)
	{
		if (config.bilinear) {
			unsigned Fx = (unsigned) (65536 * ((x + 8.0f) * xt));
			unsigned Fy = (unsigned) (65536 * ((y + 8.0f) * yt));
			unsigned xi = Fx >> 16, yi = Fy >> 16;
			Fx &= 0xffff; Fy &= 0xffff;
			unsigned x0y0, x1y0, x0y1, x1y1;
			unsigned off = (yi * size + xi) & bmask;
			x0y0 = data[off];
			x1y0 = data[(off        + 1) & bmask];
			x0y1 = data[(off + size    ) & bmask];
			x1y1 = data[(off + size + 1) & bmask];
			unsigned m0, m1, m2, m3;
			unsigned r0, g0, b0, r1, g1, b1;
			m1 = m0 = 65536 - Fy;
			m2 = 65536 - (m3=Fx);
			m0 *= m2;
			m1 *= Fx;
			m2 *= Fy;
			m3 *= Fy;
			m0 >>= 16; m1>>=16; m2 >>= 16; m3>>=16;

			b0  = m0 * (x0y0&0xff); x0y0 >>= 8;
			g0  = m0 * (x0y0&0xff); x0y0 >>= 8;
			r0  = m0 * (x0y0&0xff);
			b1  = m1 * (x1y0&0xff); x1y0 >>= 8;
			g1  = m1 * (x1y0&0xff); x1y0 >>= 8;
			r1  = m1 * (x1y0&0xff);

			b0 += m2 * (x0y1&0xff); x0y1 >>= 8;
			g0 += m2 * (x0y1&0xff); x0y1 >>= 8;
			r0 += m2 * (x0y1&0xff);
			b1 += m3 * (x1y1&0xff); x1y1 >>= 8;
			g1 += m3 * (x1y1&0xff); x1y1 >>= 8;
			r1 += m3 * (x1y1&0xff);


			b0 += b1; g0 += g1; r0 += r1;
			b0 >>= 16; g0>>=16; r0>>=16;
			return (b0 + (g0 << 8) + (r0 << 16));
			
		} else {
			unsigned Fx = (unsigned) (65536 * ((x + 8.0f) * xt));
			unsigned Fy = (unsigned) (65536 * ((y + 8.0f) * yt));
			unsigned xi = (Fx+32768) >> 16, yi = (Fy+32768) >> 16;
			return data[(yi * size + xi) & bmask];
		}
	}
};

static ProceduralBoardTexture pbt;
static LookupTexture lut;

Board::Board()
{
	if (config.procedural)
		texture = &pbt;
	else
		texture = &lut;
}

double Board::intersect(const Ray &ray, IContext &ctx)
{
	if (ray.dir[1] > -1e-9) return 1e99;
	double d = -ray.c[1] / ray.dir[1];
	ctx.ip = ray.c + ray.dir * d;
	if (fabs(ctx.ip[0]) > 8 || fabs(ctx.ip[2]) > 8) return 1e99;
	ctx.id = 42;
	ctx.object = this;
	return d * ray.dir.length();
}

unsigned Board::shade(const Ray &ray, IContext &ctx, const Light &light)
{
	unsigned color = texture->sample(ctx.ip[0], ctx.ip[2]);
	bool sh = false;
	color = ((color >> 1) & 0x7f7f7f) + ((color >> 2) & 0x3f3f3f);
	if (config.shadows && renderer.in_shadow(ctx.ip, this)) {
		color = ((color >> 2) & 0x3f3f3f3f) + ((color >> 3) & 0x1f1f1f1f);
		sh = true;
	}
	if (config.reflections) {
		Ray newray(ctx.ip, Vector(ray.dir[0], - ray.dir[1], ray.dir[2]));
		unsigned refl = renderer.raytrace(newray, this);
		if (refl) {
			float m = sh ? 0.75f : 0.375f;
			unsigned rc, gc, bc, rr, gr, br;
			bc = color & 0xff; br = refl & 0xff; color >>= 8; refl >>= 8;
			gc = color & 0xff; gr = refl & 0xff; color >>= 8; refl >>= 8;
			rc = color & 0xff; rr = refl & 0xff;
			bc = min(bc + (unsigned)(br * m), 255u);
			gc = min(gc + (unsigned)(gr * m), 255u);
			rc = min(rc + (unsigned)(rr * m), 255u);
			color = bc | (gc << 8) | (rc << 16);
		}
	}
	return color;
}

BBox Board::get_bbox(void)
{
	return BBox(Vector(-8, 0, -8), Vector(8, 0, 8));
}

int Board::get_id(void) { return 42; }

void prepare_texture(void)
{
	lut.init(pbt, 512);
}
