/***************************************************************************
 *   Copyright (C) 2007 by Veselin Georgiev                                *
 *   anrieff@mgail.com (convert to GMail)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "camera.h"
#include "config.h"
#include "matrix.h"

Camera::Camera() {init();}

void Camera::init(void)
{
	look_at.zero();
	dist = 15;
	alpha = 0.6;
	beta = .7;
	fov = 1.0;
}

void Camera::frame_init(void)
{
	cam = Vector(cos(alpha) * cos(beta), sin(beta), sin(alpha) * cos(beta));
	cam = cam * dist + look_at;
	calculate_bases();
	colinc = (w[1] - w[0]) * (1.0 / config.xres);
	rowinc = (w[2] - w[0]) * (1.0 / config.yres);
}

Vector Camera::point_rotate(Vector v)
{
	return rotation(AXIS_Y, alpha+M_PI) * (rotation(AXIS_Z, beta) * v);
}


void Camera::calculate_bases(void)
{
	double l = sqrt(2.0) * 0.5;
	double angleofview = (fov * M_PI_2) * 0.5;
	double x = tan(angleofview) * l;
	double y = x / config.aspect;
	w[0] = point_rotate(Vector(l, y, x)) + cam;
	w[1] = point_rotate(Vector(l, y, -x)) + cam;
	w[2] = point_rotate(Vector(l, -y, x)) + cam;
}

void Camera::rotate(int mx, int my)
{
	rotate_real(-mx * config.mouse_sensitivity * 0.125, my * config.mouse_sensitivity * 0.125);
}

void Camera::rotate_real(double d_alpha, double d_beta)
{
	alpha += d_alpha;
	beta  += d_beta;
	while (alpha < 0) alpha += 2 * M_PI;
	while (alpha >= 2 * M_PI) alpha -= 2 * M_PI;
	if (beta < 0) beta = 0;
	if (beta > M_PI * 0.5) beta = M_PI * 0.5;
}

void Camera::pan(int mx, int my)
{
	Vector fwd(cos(alpha), 0, sin(alpha));
	Vector left(cos(alpha + M_PI * 0.5), 0, sin(alpha + M_PI * 0.5));
	look_at -= fwd * (my * config.mouse_sensitivity);
	look_at -= left * (mx * config.mouse_sensitivity);
}

void Camera::dolly(int my)
{
	float f = my * config.mouse_sensitivity;
	f *= -0.1;
	dist = exp(log(dist) + f);
}

Ray Camera::get_screen_ray(double screen_x, double screen_y)
{
	Vector through = w[0] + colinc*screen_x + rowinc * screen_y;
	return Ray(cam, through - cam);
}

Vector Camera::row_increment(void) const
{
	return rowinc;
}

Vector Camera::col_increment(void) const
{
	return colinc;
}

static void InitMatrix(double m[3][3], const Vector&a, const Vector &b, const Vector &c) 
{
	m[0][0]=a[0]; 
	m[1][0]=b[0]; 
	m[2][0]=-c[0];
	m[0][1]=0;
	m[1][1]=b[1];
	m[2][1]=-c[1];
	m[0][2]=a[2];
	m[1][2]=b[2];
	m[2][2]=-c[2];
}


// calculates the determinant of the 3*3 matrix M.
static double MatrixRoot(double M[3][3])
{
 return (M[0][0]*M[1][1]*M[2][2] + M[1][0]*M[2][1]*M[0][2] + M[0][1]*M[1][2]*M[2][0]
        -M[0][2]*M[1][1]*M[2][0] - M[2][2]*M[0][1]*M[1][0] - M[1][2]*M[2][1]*M[0][0]);
}
#define ReplaceCol(m, col, a) m[col][0]=a[0], m[col][1]=a[1], m[col][2]=a[2]

vec2f Camera::reverse(const Vector &v)
{
	Vector a, b, c, h;
	double m[3][3];
	double Dcr, rDcr;
	double beta, gamma, delta;
		
	// init the a, b, c & h vectors
	h = cam - w[0];
	a = w[1] - w[0];
	b = w[2] - w[0];
	c.make_vector(v, cam);
	
	InitMatrix(m, a, b, c);
	Dcr = MatrixRoot(m);
	if (fabs(Dcr) < 0.01) {
		return vec2f(0,0);
	} else rDcr = 1 / Dcr;
	
	// solve the ugly equation
	ReplaceCol(m, 0, h);
	beta = MatrixRoot(m) * rDcr;
	InitMatrix(m, a, b, c);
	ReplaceCol(m, 1, h);
	gamma = MatrixRoot(m) * rDcr;
	InitMatrix(m, a, b, c);
	ReplaceCol(m, 2, h);
	delta = MatrixRoot(m) * rDcr;
	
	if (delta < 0) {
		return vec2f(0,0);
	}
	return vec2f(beta * config.xres, gamma * config.yres);
}

Camera camera;
