/***************************************************************************
 *   Copyright (C) 2007 by Veselin Georgiev                                *
 *   anrieff@mgail.com (convert to GMail)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <SDL/SDL.h>
#include "config.h"
#include "display.h"
#include "camera.h"
#include "bitmap.h"

bool Rect::has(int x, int y)
{
	return x >= x0 && x < x0 + w && y >= y0 && y < y0 + h;
}

bool Display::init(void)
{
	x = config.xres;
	y = config.yres;
	if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) < 0) {
		printf("Cannot init SDL, because the following error occured:\n%s\n",
		       SDL_GetError());
		return false;
	}
	surface = SDL_SetVideoMode(x, y, 32, config.fullscreen ? SDL_FULLSCREEN : 0);
	if (!surface) {
		printf("Cannot create drawable, because the following error occurred:\n%s\n",
		       SDL_GetError());
		return false;
	}
	buff = new unsigned [x * y];
	if (!buff) {
		printf("Cannot allocate memory for video buffer!\n");
		return false;
	}
	SDL_WM_SetCaption("got.angry?", NULL);
	clear();
	return true;
}

void Display::close(void)
{
	SDL_Quit();
	if (buff)
		delete[] buff;
	buff = NULL;
}

unsigned * Display::buffer(void) { return buff; }

void Display::update(void)
{
	memcpy(surface->pixels, buff, x * y * sizeof(buff[0]));
	SDL_Flip(surface);
}

void Display::clear(unsigned color)
{
	for (int i = 0; i < x * y; i++)
		buff[i] = color;
}

void Display::screenshot(void)
{
	Bitmap bmp(x, y);
	memcpy(bmp.data(), buff, x * y * sizeof(unsigned));
	int i = 1;
	FILE *f;
	char fn[200];
	while (1) {
		sprintf(fn, "screenshot_%04d.bmp", i);
		f = fopen(fn, "rb");
		if (f) {
			fclose(f);
			i++;
		} else {
			f = fopen(fn, "wb");
			if (!f) return;
			fclose(f);
			break;
		}
	}
	bmp.save_to_file(fn);
}

/**
 * @class GUI
 */

GUI::GUI()
{
	color = 0xffffff;
	cx = cy = 0;
	mousebuttons = 0;
	ccount = 0;
}

void GUI::sleep(int ms)
{
	SDL_Delay(ms);
}

void GUI::pause(void)
{
	SDL_Event e;
	bool she = false;
	while (!she) {
		while (SDL_PollEvent(&e)) {
			switch (e.type) {
				case SDL_QUIT:
				case SDL_KEYDOWN:
				case SDL_MOUSEBUTTONDOWN:
					she = true;
				default:
					break;
			}
			if (she) break;
		}
		SDL_Delay(50);
	}
}

void GUI::handle_inputs(bool & she)
{
	SDL_Event e;
	while (SDL_PollEvent(&e)) {
		switch (e.type) {
			case SDL_QUIT:
				she = true;
				break;
			case SDL_KEYDOWN:
			{
				switch (e.key.keysym.sym) {
					case SDLK_ESCAPE:
						she = true;
						break;
					case SDLK_f:
						config.focal_blur = 1 - config.focal_blur;
						break;
					case SDLK_q:
						config.blur_samples = 4;
						break;
					case SDLK_w:
						config.blur_samples = 12;
						break;
					case SDLK_e:
						config.blur_samples = 50;
						break;
					case SDLK_r:
						config.reflections = 1 - config.reflections;
						break;
					case SDLK_t:
						config.blur_samples = 200;
						break;
					case SDLK_s:
						config.shadows = 1 - config.shadows;
						break;
					case SDLK_a:
						config.aa = 1 - config.aa;
						break;
					case SDLK_p:
						config.procedural = 1 - config.procedural;
						break;
					case SDLK_b:
						config.bilinear = 1 - config.bilinear;
						break;
					case SDLK_F5:
						config.cel_shading = 1 - config.cel_shading;
						break;
					case SDLK_F12:
					{
						display.screenshot();
						break;
					}
					default:
						break;
				}
				break;
			}
			case SDL_MOUSEBUTTONDOWN:
			case SDL_MOUSEBUTTONUP:
			{
				int cindex = ccount - 1;
				int x = e.button.x;
				int y = e.button.y;
				while (cindex >= 0) {
					if (clickables[cindex].rect.has(x, y))
						break;
					cindex--;
				}
				int mask = 0;
				switch (e.button.button) {
					case SDL_BUTTON_LEFT:
						mask = BLEFT;
						break;
					case SDL_BUTTON_RIGHT:
						mask = BRIGHT;
						break;
					case SDL_BUTTON_MIDDLE:
						mask = BMIDDLE;
						break;
					default:
						break;
				}
				if (e.button.state == SDL_PRESSED) {
					mousebuttons |= mask;
					if (cindex > -1) {
						mousebuttons |= CLICKABLE_MASK;
					}
				} else {
					mousebuttons &= ~mask;
					mousebuttons &= ~CLICKABLE_MASK;
					if (cindex > -1)
						clickables[cindex].callback->process(x, y);
				}
				break;
			}
			case SDL_MOUSEMOTION:
			{
				int x = e.motion.xrel;
				int y = e.motion.yrel;
				if (!(mousebuttons & CLICKABLE_MASK)) {
					if (mousebuttons & BLEFT)
						camera.rotate(x, y);
					if (mousebuttons & BRIGHT)
						camera.dolly(y);
					if (mousebuttons & BMIDDLE)
						camera.pan(x, y);
				}
				break;
			}
			default:
				break;
		}
	}
}

bool GUI::cursor_inside(Rect rect)
{
	int x, y;
	SDL_GetMouseState(&x, &y);
	x -= rect.x0;
	y -= rect.y0;
	return (x >= 0 && x < rect.w && y >=0  && y < rect.h);
}

double GUI::btime(void)
{
	return SDL_GetTicks() * 0.001;
}

bool GUI::pending_events(void)
{
	return 1 == SDL_PollEvent(NULL);
}

void GUI::remove_all_clickables(void)
{
	ccount = 0;
}

void GUI::add_clickable(Rect rect, ClickCallback *callback)
{
	clickables[ccount].rect = rect;
	clickables[ccount++].callback = callback;
}

void GUI::remove_clickable(ClickCallback *callback)
{
	int i = 0;
	while (i < ccount && clickables[i].callback != callback) {
		i++;
	}
	if (i >= ccount) return;
	--ccount;
	while (i < ccount) {
		clickables[i] = clickables[i+1];
		i++;
	}
}

Display display;
GUI ui;
