/***************************************************************************
 *   Copyright (C) 2007 by Veselin Georgiev                                *
 *   anrieff@mgail.com (convert to GMail)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <string.h>
#include <stdlib.h>
#include "config.h"
#include "display.h"
#include "render.h"
#include "camera.h"
#include "board.h"
#include "pawn.h"
#include "vista.h"
#include "threads.h"
#include "array.h"
#include "vector2f.h"


const int COLOR_DIFF_THRESH = 250u;
const double APERTURE = 0.5;

static inline unsigned lsb(unsigned x)
{
	int r = 0;
	if (!(x & 0xffff)) { r += 16; x >>= 16; }
	if (!(x & 0x00ff)) { r += 8; x >>= 8; }
	if (!(x & 0x000f)) { r += 4; x >>= 4; }
	if (!(x & 0x0003)) { r += 2; x >>= 2; }
	if (!(x & 0x0001)) r++;
	return r;
}

static inline unsigned popcount(unsigned x)
{
	x = (x & 0x55555555) + ((x & 0xaaaaaaaa) >> 1);
	x = (x & 0x33333333) + ((x & 0xcccccccc) >> 2);
	x = (x & 0x0f0f0f0f) + ((x & 0xf0f0f0f0) >> 4);
	x = (x & 0x00ff00ff) + ((x & 0xff00ff00) >> 8);
	x = (x & 0x0000ffff) + ((x & 0xffff0000) >> 16);
	return x;
}

unsigned Renderer::raytrace(const Ray & ray, Object* last_object, unsigned check_mask, int buff_index)
{
	if (config.focal_blur && !last_object) {
		return raytrace_focal_blur(ray);
	}

	check_mask &= (1 << objnum) - 1;
	double mind = 1e99;
	IContext best, ctx;
	Object *z = NULL;
	int i;
	while (check_mask) {
		i = lsb(check_mask);
		check_mask ^= (1 << i);
		if (objects[i] == last_object) continue;
		//
		double d = objects[i]->intersect(ray, ctx);
		if (d < mind) {
			mind = d;
			best = ctx;
			z = objects[i];
		}
	}
	if (z) {
		if (buff_index > -1) {
			idsbuff[buff_index] = best.id;
		}
		return z->shade(ray, best, light);
	}
	return 0;
}

void sample_lens_point(double &x, double &y)
{
	double angle = dice_random.drand() * 2 * M_PI;
	double r = dice_random.drand();
	r = sqrt(r);
	x = cos(angle) * r;
	y = sin(angle) * r;
}


unsigned Renderer::raytrace_focal_blur(const Ray & ray)
{
	double mind;
	IContext ctx;
	//Vector t = ray.c + ray.dir * (focal_dist / ray.dir.length());
	double xd = ray.c * fpnorm + fpd;
	double xdd = ray.dir * fpnorm;
	Vector t = ray.c + ray.dir * (-xd / xdd);
	int r = 0, g = 0, b = 0;
	int n = config.blur_samples;
	for (int smp = 0; smp < n; smp++) {
		double x, y;
		sample_lens_point(x, y);
		Vector newcam = ray.c + fp0 * x + fp1 * y;
		Ray newray(newcam, t - newcam);
		Object *z = NULL;
		IContext best;
		mind = 1e99;
		for (int i = 0; i < objnum; i++) {
			double d = objects[i]->intersect(newray, ctx);
			if (d < mind) {
				mind = d;
				z = objects[i];
				best = ctx;
			}
		}
		if (z) {
			unsigned res = z->shade(newray, best, light);
			b += res & 0xff; res >>= 8;
			g += res & 0xff; res >>= 8;
			r += res & 0xff;
		}
	}
	return (b / n) | ((g / n) << 8) | ((r / n) << 16);
}

bool Renderer::too_much_color_difference(unsigned a0, unsigned a1, unsigned a2, unsigned a3)
{
	int diff = 0;
	int t[4] = {a0, a1, a2, a3};
	for (int k = 0; k < 3; k++) {
		int temp = 0;
		for (int i = 0; i < 4; i++) {
			temp += t[i] & 0xff;
		}
		for (int i = 0; i < 4; i++) {
			diff += abs(4*((int) (t[i] & 0xff)) - temp);
			if (diff > COLOR_DIFF_THRESH) return true;
			t[i] >>= 8;
		}
	}
	return false;
}

static inline unsigned expand(unsigned x)
{
	return (x & 0xff) | ((x & 0xff00) << 2) | ((x & 0xff0000) << 4);
}

static inline unsigned shrink(unsigned x)
{
	return ((x >> 2) & 0xff) | ((x >> 4) & 0xff00) | ((x >> 6) & 0xff0000);
}

unsigned Renderer::rec_antialias(unsigned mask,
		double x0, double y0, double x1, double y1,
		unsigned c0, unsigned c1, unsigned c2, unsigned c3, int level)
{
	const unsigned _ = 0xffffffff;
	if (_ == c0)
		c0 = raytrace(camera.get_screen_ray(x0, y0), NULL, mask);
	if (_ == c1)
		c1 = raytrace(camera.get_screen_ray(x1, y0), NULL, mask);
	if (_ == c2)
		c2 = raytrace(camera.get_screen_ray(x0, y1), NULL, mask);
	if (_ == c3)
		c3 = raytrace(camera.get_screen_ray(x1, y1), NULL, mask);
	int r = 0, g = 0, b = 0;
	b = (c0 & 0xff) + (c1 & 0xff) + (c2 & 0xff) + (c3 & 0xff);
	g = ((c0 >> 8) & 0xff) + ((c1 >> 8) & 0xff) + ((c2 >> 8) & 0xff) + ((c3 >> 8) & 0xff);
	r = ((c0 >> 16) & 0xff) + ((c1 >> 16) & 0xff) + ((c2 >> 16) & 0xff) + ((c3 >> 16) & 0xff);
	if (level > 0)
		return (unsigned) ((b >> 2) | ((g << 6) & 0xff00) | ((r << 14) & 0xff0000));
	int diff = abs(b - 4*((int) (c0 & 0xff)))
		+ abs(b - 4*((int) (c1 & 0xff)))
		+ abs(b - 4*((int) (c2 & 0xff)))
		+ abs(b - 4*((int) (c3 & 0xff)))
		+ abs(g - 4*((int) ((c0 >> 8) & 0xff)))
		+ abs(g - 4*((int) ((c1 >> 8) & 0xff)))
		+ abs(g - 4*((int) ((c2 >> 8) & 0xff)))
		+ abs(g - 4*((int) ((c3 >> 8) & 0xff)))
		+ abs(r - 4*((int) ((c0 >> 16) & 0xff)))
		+ abs(r - 4*((int) ((c1 >> 16) & 0xff)))
		+ abs(r - 4*((int) ((c2 >> 16) & 0xff)))
		+ abs(r - 4*((int) ((c3 >> 16) & 0xff)));
	if (diff <= COLOR_DIFF_THRESH) {
		return (unsigned) ((b >> 2) | ((g << 6) & 0xff00) | ((r << 14) & 0xff0000));
	}
	double xm = (x0 + x1) * 0.5;
	double ym = (y0 + y1) * 0.5;
	int l1 = level + 1;
	unsigned r0 = rec_antialias(mask, x0, y0, xm, ym, c0, _, _, _, l1);
	unsigned r1 = rec_antialias(mask, xm, y0, x1, ym, _, c1, _, _, l1);
	unsigned r2 = rec_antialias(mask, x0, ym, xm, y1, _, _, c2, _, l1);
	unsigned r3 = rec_antialias(mask, xm, ym, x1, y1, _, _, _, c3, l1);
	return shrink(expand(r0) + expand(r1) + expand(r2) + expand(r3)/* + 0x200802*/);
}

void Renderer::render_mt_spawner(void)
{
	class Temp : public Parallel {
		void entry(int, int) {
			renderer.render_mt();
		}
	} tmp;
	pool->run(&tmp, get_processor_count());
}

#define BLOCK_SIZE 64
void Renderer::render_mt(void)
{
	int wx, wy;
	wx = (config.xres - 1) / BLOCK_SIZE + 1;
	wy = (config.yres - 1) / BLOCK_SIZE + 1;
	int all = wx * wy, current = 0;
	unsigned *b = display.buffer();
	unsigned *v = vista;
	Vector rowinc = camera.row_increment();
	Vector colinc = camera.col_increment();

	while ((current = (*counter)++) < all) {
		int x0 = (current % wx) * BLOCK_SIZE;
		int y0 = (current / wx) * BLOCK_SIZE;
		int w = x0 + BLOCK_SIZE <= config.xres ? BLOCK_SIZE : config.xres - x0;
		int h = y0 + BLOCK_SIZE <= config.yres ? BLOCK_SIZE : config.yres - y0;
		int wr = (config.aa && x0 + w < config.xres) ? w + 1 : w;
		int hr = (config.aa && y0 + h < config.yres) ? h + 1 : h;
		Ray ry = camera.get_screen_ray(x0, y0);
		Ray r = ry;
		for (int y = 0; y < hr; y++) {
			int index = x0 + (y0+y) * config.xres;
			r.dir = ry.dir;
			for (int x = 0; x < wr; x++, r.dir += colinc, index++) {
				if (!v[index] && !config.focal_blur) {
					b[index] = 0; continue;
				}
				b[index] = raytrace(r, NULL, v[index], index);
			}
			ry.dir += rowinc;
		}
		if (!config.aa || config.focal_blur) continue;
		--wr, --hr;
		for (int y = 0; y < hr; y++) {
			int index = x0 + (y+y0) * config.xres;
			for (int x = 0; x < wr; x++, index++) {
				if (!too_much_color_difference(
					b[index], b[index+1],
					b[index+config.xres], b[index+config.xres+1])) continue;
				const unsigned _ = 0xffffffff;
				unsigned mask = v[index] | v[index+1] | v[index+config.xres] | v[index+config.xres+1];
				b[index] = rec_antialias(mask, x + x0 - 0.5, y + y0 - 0.5, x + x0 + 0.5, y + y0 + 0.5, _, _, _, _);
			}
			
		} 
	}
}

void Renderer::render(void)
{
	camera.frame_init();
	prepare_vista_buffer();
	prepare_glow_buffer();
	if (config.shadows)
		prepare_shadowed_areas();
	rowinc = camera.row_increment();
	colinc = camera.col_increment();
	if (config.focal_blur)
		prepare_focal_dist();
	if (get_processor_count()>1) {
		*counter = 0;
		render_mt_spawner();
	} else {
		unsigned *b = display.buffer();
		unsigned *v = vista;
		Ray bray = camera.get_screen_ray(0, 0);
		for (int y = 0; y < config.yres; y++) {
			Ray r = bray;
			for (int x = 0; x < config.xres; x++, b++, v++, r.dir += colinc) {
				if (!(*v) && !config.focal_blur) {
					*b = 0; continue;
				}
				
				*b = raytrace(r, NULL, *v, y * config.xres + x);
				
/*				int t = popcount(*v) * 32;
				*b = t | (t << 8) | (t << 16);*/
			}
			bray.dir += rowinc;
		}
		
		if (config.aa && !config.focal_blur && !config.cel_shading) {
			unsigned char *X = idsbuff;
			unsigned *b = display.buffer();
			unsigned *v = vista;
			for (int y = 0; y < config.yres-1; y++) {
				for (int x = 0; x < config.xres-1; x++, b++, v++, X++) {
					if (!too_much_color_difference(b[0], b[1], b[config.xres], b[config.xres+1])) continue;
					//if (!(X[0] != X[1] || X[0] != X[config.xres] || X[0] != X[config.xres+1])) continue;
					const unsigned _ = 0xffffffff;
					unsigned mask = v[0] | v[1] | v[config.xres] | v[config.xres+1];
					b[0] = rec_antialias(mask,
						x - 0.5, y - 0.5,
						x + 0.5, y + 0.5,
						_, _, _, _);
				}
				b++; v++; X++;
			}
		}
	}
	if (config.cel_shading && !config.focal_blur) {
		apply_cel_shading();
	}
	
	if (glow_ids_count) {
		prepare_blur_buff();
		blur();
		apply_glow();
	}
}

void Renderer::reinit_scene(void)
{
	clear_objects();
	light.pos = Vector(10, 9, 14);
	light.intensity = 5;
	objects[objnum++] = new Board;
	
}

Renderer::Renderer() {
	memset(objects, 0, sizeof(objects)); objnum = 0; vista = NULL;
	pool = NULL;
	counter = NULL;
	if (get_processor_count() > 1) {
		pool = new ThreadPool;
		counter = new InterlockedInt;
	}
}


Renderer::~Renderer()
{
	clear_objects();
	if (vista)
		delete [] vista;
	vista = NULL;
	if (idsbuff)
		delete [] idsbuff;
	idsbuff = NULL;
	if (blurbuff)
		delete [] blurbuff;
	blurbuff = NULL;
	if (temp)
		delete [] temp;
	temp = NULL;
	if (pool)
		delete pool;
	pool = NULL;
	if (counter)
		delete counter;
	counter = NULL;
}

void Renderer::clear_objects(void)
{
	for (int i = 0; i < objnum; i++) {
		if (objects[i] != NULL)
			delete objects[i];
		objects[i] = NULL;
	}
	objnum = 0;
}

void Renderer::add_pawn(Vector where, unsigned color, int id)
{
	if (color < 4)
		color = player_colors[color];
	objects[objnum++] = new Pawn(where, color, id);
}


void Renderer::prepare_vista_buffer(void)
{
	if (!vista)
		vista = new unsigned[config.xres * config.yres];
	memset(vista, 0, sizeof(unsigned) * config.xres * config.yres);
	draw_vista(vista, objects, objnum);
}

void Renderer::prepare_glow_buffer(void)
{
	if (!idsbuff)
		idsbuff = new unsigned char[config.xres * config.yres];
	if (!blurbuff)
		blurbuff = new unsigned char[config.xres * config.yres];
	memset(idsbuff, 0xff, config.xres * config.yres);
}


Vector interpolate(const AnimationEntry &e, double x)
{
	Vector v0 = e.start_pos;
	Vector v1 = e.end_pos;
	Vector vd = v1 - v0;
	double scale = vd.length();
	Vector res = v0 + vd * x + Vector(0, sin(x*M_PI) * scale * 0.5, 0);
	return res;
}

void Renderer::animate(AnimationEntry *entries, int num_entries, bool & she, Callback &cb)
{
	double max_time = 0;
	for (int i = 0; i < num_entries; i++) {
		max_time = max(max_time, (double)entries[i].end_time);
	}
	double start = ui.btime();
	double t;
	while ((t = (ui.btime() - start)*config.speed) < max_time) {
		for (int i = 0; i < num_entries; i++) {
			AnimationEntry &e = entries[i];
			if (e.start_time > t || t >= e.end_time) continue;
			double x = (t - e.start_time) / (e.end_time - e.start_time);
			for (int i = 0; i < objnum; i++)
				if (objects[i]->get_id() == e.id) {
					objects[i]->move(e.start_pos, e.end_pos, x);
				}
		}
		render();
		cb.process();
		display.update();
		ui.handle_inputs(she);
	}
}

void Renderer::enable_glow(int a[], int n)
{
	glow_ids_count = n;
	for (int i = 0; i < n; i++)
		glow_ids[i] = a[i];
}

void Renderer::disable_glow(void)
{
	glow_ids_count = 0;
}

int Renderer::get_id_at(int x, int y)
{
	if (!idsbuff) return 0xff;
	return idsbuff[y * config.xres + x];
}

void Renderer::prepare_blur_buff(void)
{
	int size = config.xres * config.yres;
	for (int i = 0; i < size; i++) {
		blurbuff[i] = 0;
		for (int j = 0; j < glow_ids_count; j++) {
			if (glow_ids[j] == idsbuff[i]) {
				blurbuff[i] = 0xff;
				break;
			}
		}
	}
}

void Renderer::blur(void)
{
	int resx = config.xres;
	int resy = config.yres;
	float coeff = 0.2f;
	unsigned I = (unsigned) (65535 * coeff);
	unsigned Im1 = 65535 - I;
	for (int row = 0; row < resy; row++) {
		unsigned char *sr = blurbuff  + row * resx;
		unsigned char *dr = blurbuff + row * resx;
		for (int x = 1; x < resx; x++) {
			dr[x] = (unsigned char) ((sr[x] * I + dr[x-1] * Im1)>>16);
		}
		for (int x = resx - 2; x >= 0; x--) {
			dr[x] = ((dr[x] * I + dr[x+1] * Im1)>>16);
		}
	}
	for (int col = 0; col < resx; col++) {
		unsigned char * p = blurbuff + col;
		for (int y = 1; y < resy; y ++ ){
			p[y * resx] = ((p[y * resx] * I + p[(y-1) * resx] * Im1) >> 16);
		}
		for (int y = resy - 2; y >= 0; y--) {
			p[y * resx] = ((p[y * resx] * I + p[(y+1) * resx] * Im1) >> 16);
		}
	}
}

void Renderer::apply_glow(void)
{
	static int ids_colors[256] = {0};
	if (ids_colors[0] == 0) {
		for (int i = 0; i < 256; i++)
			ids_colors[i] = rand() & 0xffffff;
	}
	int size = config.xres * config.yres;
	unsigned *b = display.buffer();
	for (int i = 0; i < size; i++) {
		bool flag = false;
		for (int j = 0; j < glow_ids_count; j++) {
			if (idsbuff[i] == glow_ids[j]) {
				flag = true;
				break;
			}
		}
		if (!flag && blurbuff[i]) {
			float intensity = blurbuff[i] / 128.0f;
			if (intensity > 1.0f) intensity = 1.0f;
			b[i] = mul(b[i], 1-intensity) + mul(0xff00, intensity);
		}
		
		//b[i] = ids_colors[idsbuff[i]];
	}
}

bool Renderer::in_shadow(const Vector & v, Object * last_object)
{
	if (!shadowed_areas[(int) (v[0] + 16)][(int) (v[2] + 16)])
		return false;
	Ray r(v, light.pos - v);
	IContext ctx;
	for (int i = 0; i < objnum; i++) if (objects[i] != last_object) {
		if (objects[i]->intersect(r, ctx) < 1e99)
			return true;
	}
	return false;
}

static inline float face(const vec2f& a, const vec2f& b, const vec2f& c) 
{
	return +a[0] * b[1] + b[0] * c[1] + c[0] * a[1]
	       -c[0] * b[1] - b[0] * a[1] - a[0] * c[1];
}

Array<vec2f> convex_hull(Array<vec2f> input)
{
	sort(&input[0], input.size());
	
	Array<vec2f> l, r;
	for (int i = 0; i < 2; i++) {
		l += input[i];
		r += input[i];
	}
	
	for (int i = 2; i < input.size(); i++) {
		l += input[i];
		r += input[i];
		while (l.size() > 2 && face(l[l.size()-3], l[l.size()-2], l[l.size()-1]) > -1e-6) {
			l.erase(l.size()-2);
		}
		while (r.size() > 2 && face(r[r.size()-3], r[r.size()-2], r[r.size()-1]) < +1e-6) {
			r.erase(r.size()-2);
		}
	}
	
	Array<vec2f> ret;
	for (int i = 0; i < r.size(); i++)
		ret += r[i];
	for (int i = l.size() - 2; i > 0; i--)
		ret += l[i];
	return ret;
}

float xarea(const Array<vec2f> &poly, const vec2f &p)
{
	float res = 0;
	for (int i = 0; i < poly.size(); i++) {
		res += fabs(face(p, poly[i], poly[(i+1) % poly.size()]));
	}
	return res;
}

bool inside(const vec2f& p, const Array<vec2f> & poly)
{
	float a = xarea(poly, poly[0]);
	float b = xarea(poly, p);
	return fabs(a-b) < 1e-5;
}

void Renderer::prepare_shadowed_areas(void)
{
	memset(shadowed_areas, 0, sizeof(shadowed_areas));
	for (int x = -8; x < 8; x++) {
		for (int y = -8; y < 8; y++) {
			bool flag = false;
			Array<vec2f> square;
			square += vec2f(x, y);
			square += vec2f(x, y+1);
			square += vec2f(x+1, y+1);
			square += vec2f(x+1, y);

			for (int i = 0; i < objnum && !flag; i++)
				if (objects[i]->get_id() != 42) { //skip the board
					BBox bbox = objects[i]->get_bbox();
					Vector v[8];
					v[0] = bbox.minv;
					v[7] = bbox.maxv;
					for (int l = 1; l < 7; l++) {
						v[l] = v[0];
						for (int k = 0; k < 3; k++)
							if ((1 << k) & l) v[l][k] = v[7][k];
					}
					Array<vec2f> in;
					Vector p0;
					for (int l = 0; l < 8; l++) {
						Vector t = v[l] - light.pos;
						p0 = v[l] + t*(v[l][1] / (light.pos[1] - v[l][1]));
						in += vec2f(p0[0], p0[2]);
					}
					Array<vec2f> out = convex_hull(in);
					for (int i = 0; i < out.size(); i++) {
						if (inside(out[i], square)) {
							flag = true;
							break;
						}
					}
					if (flag) break;
					for (int i = 0; i < square.size(); i++) {
						if (inside(square[i], out)) {
							flag = true;
							break;
						} 
					}
				}
			shadowed_areas[x + 16][y + 16] = flag; 
		}
	}
}

void Renderer::prepare_focal_dist(void)
{
	IContext ctx;
	double mind = 1e99;
	Ray ray = camera.get_screen_ray(config.xres/2, config.yres/2);
	for (int i = 0; i < objnum; i++) {
		mind = min(mind, objects[i]->intersect(ray, ctx));
	}
	if (mind > 1e10) mind = 1000;
	focal_dist = mind;
	fp_center = ray.c + ray.dir * (mind / ray.dir.length());
	fp0 = perpendicular(ray.dir);
	fp1 = ray.dir ^ fp0;
	fp0.make_length(APERTURE);
	fp1.make_length(APERTURE);
	fpnorm = fp0 ^ fp1;
	fpd = - (fpnorm * fp_center);
}

void Renderer::apply_cel_shading(void)
{
	if (!temp)
		temp = new unsigned [config.xres * config.yres];
	memcpy(temp, display.buffer(), config.xres * config.yres * sizeof(unsigned));
	unsigned * src = temp;
	unsigned * dest = display.buffer();
	unsigned char * v = idsbuff;
	int x = config.xres, y = config.yres;
	for (int j = 0; j < y; j++) {
		for (int i = 0; i < x; i++, src++, dest++, v++) {
			if (!i || !j || i == x -1 || j == y - 1) {
				*dest = 0;
				continue;
			}
			unsigned t = v[-x-1];
			if (t != v[-x] || t != v[-x-1] || t != v[-1] || t != v[0] || t != v[1] ||
			    t != v[x-1] || t != v[x] || t != v[x+1]) {
			    	*dest = 0;
			} else {
				t = *src;
				*dest = (((t >> 6) & 0x030303) << 6)+0x202020;
			}
		}
	}
}

Renderer renderer;
