/***************************************************************************
 *   Copyright (C) 2007 by Veselin Georgiev                                *
 *   anrieff@mgail.com (convert to GMail)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __RENDER_H__
#define __RENDER_H__

#include "ray.h"
#include "matrix.h"
#include "utils.h"

class Object;

struct IContext { // Intersection Context
	Vector ip;// intersection point
	Object *object;
	int id;
	char extra[32]; // extra padding
};

struct Light {
	Vector pos;
	float intensity;
};

struct BBox {
	Vector minv, maxv;
	BBox(){}
	BBox(const Vector &_minv, const Vector &_maxv) : minv(_minv), maxv(_maxv) {}
};

/**
 * @class Object - generic class for representing intersectable and shadeable objects
 * must implement intersect() and shade() methods
 */
class Object {
public:
	/**
	 * intersect()s the object with the given ray.
	 * If the ray doesn't intersect the object, 1e99 must be returned
	 * If the intersection occurs, the exact distance from the camera origin
	 * to the intersection must be returned, and the context structure
	 * filled accordingly
	 */
	virtual double intersect(const Ray &ray, IContext& context) = 0;
	
	/**
	 * executes the shader for the object at the intersection point,
	 * as stored in the context.
	 * Returns: a valid color
	 */
	virtual unsigned shade(const Ray& ray, IContext& context, const Light & light) = 0;
	
	/**
	 * gets the bounding box of the object
	 */
	virtual BBox get_bbox(void) = 0;
	
	/// get object's ID
	virtual int get_id(void) = 0;
	
	/// move the object's center to the new location
	virtual void move(Vector start_pos, Vector end_pos, double x) {}
	virtual ~Object(){}
};

struct AnimationEntry {
	int id;
	int start_time, end_time;
	Vector start_pos, end_pos;
};

class ThreadPool;
class InterlockedInt;

class Renderer {
	Light light;
	Object * objects[20];
	unsigned *vista, *temp;
	unsigned char *idsbuff, *blurbuff;
	int objnum;
	ThreadPool *pool;
	InterlockedInt *counter;
	Vector rowinc, colinc;
	Vector xcam;
	double focal_dist, fpd;
	Vector fp0, fp1, fp_center, fpnorm;
	
	
	void prepare_vista_buffer(void);
	void prepare_glow_buffer(void);
	void prepare_blur_buff();
	void prepare_shadowed_areas(void);
	void prepare_focal_dist(void);
	void blur();
	void apply_glow();
	void apply_cel_shading(void);
	bool too_much_color_difference(unsigned a0, unsigned a1, unsigned a2, unsigned a3);
	unsigned rec_antialias(unsigned, double, double, double, double, unsigned, unsigned, unsigned, unsigned, int level = 0);
	
	
	int glow_ids[20];
	int glow_ids_count;
public:
	bool shadowed_areas[32][32];
	Renderer();
	~Renderer();
	void reinit_scene(void);
	void add_pawn(Vector where, unsigned color, int id);
	void render(void);
	void render_mt(void);
	void render_mt_spawner(void);
	void clear_objects(void);
	unsigned raytrace(const Ray & ray, Object * last_object, unsigned mask = 0xffffffff, int buff_index = -1);
	unsigned raytrace_focal_blur(const Ray & ray);
	void animate(AnimationEntry *entries, int num_entries, bool & she, Callback & per_frame_callback);
	void enable_glow(int a[], int n);
	void disable_glow(void);
	int get_id_at(int x, int y);
	bool in_shadow(const Vector & point, Object * last_object);
};


const unsigned player_colors[4] = {
	0xff2222,
	0xefef12,
	0x15ce15,
	0x2525ff,
};


const unsigned bgnd_color = 0x777788;
const unsigned border_color = 0x111111;
extern Renderer renderer;

#endif // __RENDER_H__
