/***************************************************************************
 *   Copyright (C) 2007 by Veselin Georgiev                                *
 *   anrieff@mgail.com (convert to GMail)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef __VECTOR3_FRACT_H__
#define __VECTOR3_FRACT_H__

#include <stdio.h>
#include <math.h>

#ifndef sqr
#define sqr(x) ((x)*(x))
#endif

/* = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = */
/**
 * @class Vector
 * @short 3D Vector with the usual operations
 *
 * Constructors:
 * [double, double, double]
 * [double*]
 * [Vector]
 *
 * `+' - adds two vectors
 * `-' - substracts two vectors
 * `*' of the form <vector>*<scalar> - scales a vector
 * `*' of the form <vector>*<vector> - Dot (scalar) product
 * `^' - Vector product
 * +=, -= - accumulation
 * ==  - true if vector are EXACTLY identical (no threshold comparison is done)
 *
 * Methods:
 * @method scale     - scales a vector by a scalar (result is not lost)
 * @method norm      - normalizes a vector to unitary
 * @method zero      - makes a vector zero
 * @method length    - returns the length
 * @method lengthSqr - returns the squared length
 * @method distto    - returns the distance to the destination
 *	@param vec - the destination vector
 * @method make_vector - makes a vector between the points A and B
 *	@param A  - first (destination) vector
 *	@param B  - second (source) vector
 * @method macc      - Multiply and accumulate;
 * 	@param start
 *	@param ray
 *	@param scale
 *	@note function: *this = start + ray * scale
 * @method add2      - assigns the sum of two vectors to *this
 * @method add3      - assigns the sum of three vectors to *this
 * @method inverse   - component-wise 1/v[i]
 */
/* = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = */

#define SCALE v[0]*=scale,v[1]*=scale,v[2]*=scale
#define vreal double
class Vector {
public:
	union {
		vreal v[4];
		struct {vreal x, y, z, a;};
	};

	Vector() {}
	Vector(const vreal& x1, const vreal& y1, const vreal& z1):x(x1), y(y1), z(z1) {}
	Vector(vreal data[3]) : x(data[0]), y(data[1]), z(data[2]) {}
	Vector(const Vector &r) : x(r.x), y(r.y), z(r.z) {}
	Vector& operator =(const Vector& r) {
		v[0] = r.v[0];
		v[1] = r.v[1];
		v[2] = r.v[2];
		return *this;
	}

	/// returns true if vectors are IDENTICAL (bitwize)
	inline bool operator == (const Vector &r) const;
	inline bool operator != (const Vector &r) const;
	
	/// returns trie if vectors are roughly the same
	inline bool like(const Vector & r, double compare_factor = 1e-9) const;

	/// calculates the vector from point `src' to point `dst'
	inline void make_vector(const Vector & dst, const Vector & src);
	/// calculates the vector: start + scale * ray
	inline void macc(const Vector & start, const Vector & ray, const vreal & scale);

	/// normalizes a non-null vector to unitary
	inline void norm();
	/// nullifies a vector
	inline void zero();
	/// returns the length of the vector
	inline vreal length() const;
	/// returns the squared length
	inline vreal lengthSqr() const;
	/// multiplies by a scalar
	inline Vector & scale(vreal const &);

	inline void add2(const Vector & x, const Vector & y);
	inline void add3(const Vector & x, const Vector & y, const Vector & z);
	inline void make_length(const vreal & new_length);

	/// addition
	inline const Vector operator +(const Vector &) const;
	/// substraction
	inline const Vector operator -(const Vector &) const;
	/// addition
	inline Vector & operator +=(const Vector &);
	/// substraction
	inline Vector & operator -=(const Vector &);

	/// dot product
	inline vreal operator * (const Vector &) const;
	/// multiplication by a scalar
	inline Vector operator *(vreal const & ) const;
	/// cross product
	inline Vector operator ^ (const Vector &) const;
	inline const vreal& operator [] (const unsigned index) const;
	inline vreal & operator [] (const unsigned index);
	inline vreal distto(const Vector &) const;
	inline void inverse(void);
	inline void print() const;
	inline bool is_finite() const;
};

inline void Vector::make_vector(const Vector & dst, const Vector & src)
{
	v[0] = dst.v[0] - src.v[0];
	v[1] = dst.v[1] - src.v[1];
	v[2] = dst.v[2] - src.v[2];
}

inline void Vector::macc(const Vector & start, const Vector & ray, const vreal & scale)
{
	v[0] = start.v[0] + scale * ray.v[0];
	v[1] = start.v[1] + scale * ray.v[1];
	v[2] = start.v[2] + scale * ray.v[2];
}

inline void Vector::norm()
{
	vreal scale = 1.0/sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);
	SCALE;
}

inline void Vector::zero()
{
	v[0] = 0.0;
	v[1] = 0.0;
	v[2] = 0.0;
}

inline Vector & Vector::scale(vreal const &  scale)
{
	SCALE;
	return *this;
}

inline vreal Vector::length() const
{
	return sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);
}

inline vreal Vector::lengthSqr() const
{
	return v[0]*v[0]+v[1]*v[1]+v[2]*v[2];
}

inline void Vector::add2(const Vector & x, const Vector & y)
{
	v[0] = x.v[0] + y.v[0];
	v[1] = x.v[1] + y.v[1];
	v[2] = x.v[2] + y.v[2];
}

inline void Vector::add3(const Vector & x, const Vector & y, const Vector & z)
{
	v[0] = x.v[0] + y.v[0] + z.v[0];
	v[1] = x.v[1] + y.v[1] + z.v[1];
	v[2] = x.v[2] + y.v[2] + z.v[2];
}

inline void Vector::make_length(const vreal & new_length)
{
	vreal scale = new_length/sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);
	SCALE;
}

inline const Vector  Vector::operator + (const Vector & r) const
{
	return Vector (
			v[0] + r.v[0],
	v[1] + r.v[1],
	v[2] + r.v[2]
		      );
}

inline const Vector  Vector::operator - (const Vector & r) const
{
	return Vector(v[0] - r.v[0],
		      v[1] - r.v[1],
		      v[2] - r.v[2]);
}

inline vreal Vector::operator * (const Vector & r) const
{
	return v[0]*r.v[0] + v[1]*r.v[1] + v[2]*r.v[2];
}

inline Vector Vector::operator *(vreal const &  r) const
{
	return Vector(v[0]*r, v[1]*r, v[2]*r);
}

inline Vector Vector::operator ^ (const Vector & r) const
{
	return Vector(v[1]*r.v[2] - v[2]*r.v[1],
		      v[2]*r.v[0] - v[0]*r.v[2],
		      v[0]*r.v[1] - v[1]*r.v[0]
		     );
}

inline Vector & Vector::operator += (const Vector & r)
{
	v[0] += r.v[0];
	v[1] += r.v[1];
	v[2] += r.v[2];
	return *this;
}

inline Vector & Vector::operator -= (const Vector & r)
{
	v[0] -= r.v[0];
	v[1] -= r.v[1];
	v[2] -= r.v[2];
	return *this;
}
inline const vreal& Vector::operator [] (const unsigned index) const
{
	return v[index];
}

inline vreal & Vector::operator [] (const unsigned index)
{
	return v[index];
}

inline vreal Vector::distto(const Vector & r) const
{
	return sqrt(sqr(r.v[0]-v[0]) + sqr(r.v[1]-v[1]) + sqr(r.v[2]-v[2]));
}

inline void Vector::inverse(void)
{
	if (fabs(v[0]) >= 1e-12) v[0] = 1.0/v[0];
	if (fabs(v[1]) >= 1e-12) v[1] = 1.0/v[1];
	if (fabs(v[2]) >= 1e-12) v[2] = 1.0/v[2];
}

inline bool Vector::operator == (const Vector & r) const
{
	return v[0] == r.v[0] && v[1] == r.v[1] && v[2] == r.v[2];
}

inline bool Vector::operator != (const Vector & r) const
{
	return v[0] != r.v[0] || v[1] != r.v[1] || v[2] != r.v[2];
}

#define equal(a,b,c) (fabs((a)-(b))<(c))

inline bool Vector::like(const Vector &r, double compare_factor) const
{
	return 	equal(v[0], r.v[0], compare_factor) && 
			equal(v[1], r.v[1], compare_factor) && 
			equal(v[2], r.v[2], compare_factor);
}

inline void Vector::print() const
{
	printf("(%.4lf, %.4lf, %.4lf)\n", x, y, z);
}

#ifdef _MSC_VER
#include <float.h>
#define finite _finite
#endif

inline bool Vector::is_finite() const
{
	return finite(v[0]) && finite(v[1]) && finite(v[2]);
}

#endif // __VECTOR3_FRACT_H__
