/***************************************************************************
 *   Copyright (C) 2007 by Veselin Georgiev                                *
 *   anrieff@mgail.com (convert to GMail)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "config.h"
#include "camera.h"
#include "vista.h"
#include "vector2f.h"


/**
 * @struct	ClipRes
 * @brief	Defines a result from FrustrumClipper::clip function
 * 
 * @param n	- Number of points in the resulting poly (may be 0 -
 * 		  this means the triangle is out of the frustrum)
 * @param v	- Vertices of the resulting poly
 * @param bary	- The barycentric coordinates of the resulting poly,
 *		  relative to the original triangle
*/
#define MAX_CLIP_RESULT 16
struct ClipRes {
	int n;
	Vector v[MAX_CLIP_RESULT];
	Vector bary[MAX_CLIP_RESULT];
};

/**
 * @class	FrustrumClipper
 * @date	2006-06-08
 * @author	Veselin Georgiev
 * @brief	Intersects and clips triangles against a view frustrum
*/ 
class FrustrumClipper {
	double D[4];
	Vector n[4];
	int defeval;
	
	void form_plane(const Vector &a, const Vector &b, const Vector &c, int index);
	double eval(const Vector &v, int which = -1);
public:
	/**
	 * Constructor:
	 * @param camera	- The location of the camera
	 * @param upleft	- a point that lies along the upleft corner ray of the view frustrum
	 * @param xdir		- X direction (upleft + xdir must be the upright corner of the frustrum)
	 * @param ydir		- Y direction (upleft + ydir must be the downleft corner of the frustrum)
	 */ 
	FrustrumClipper(const Vector &camera, const Vector& upleft, const Vector& xdir, const Vector &ydir);
	
	/**
	 * Performs clipping of the given triangle, with the specified planes
	 * @param clipres	Struct to store result
	 * @param trio		Array of 3 Vectors (the triangle itself)
	 * @param which		A bit mask of the view planes to clip against (may be ORed together):
	 * 	1	- up plane
	 * 	2	- left plane
	 * 	4	- right plane
	 * 	8	- down plane
	 * 	(the default, 15, is "clip against all planes")
	*/ 
	void clip(ClipRes& clipres, Vector trio[], int which = 15);
};

/** ** ** ** ** ** ** ** ** ** ** ** **
 * ** ** @class FrustrumClipper ** ** **
 * *** ** ** ** ** ** ** ** ** ** ** **
 */

const Vector imatrix[3] = {
	Vector(1.0, 0.0, 0.0),
	Vector(0.0, 1.0, 0.0),
	Vector(0.0, 0.0, 1.0)
};

FrustrumClipper::FrustrumClipper(const Vector& camera, const Vector& ul, const Vector& x, const Vector &y)
{
	Vector e[4];
	e[0] = ul;
	e[1] = ul + x;
	e[2] = ul + y;
	e[3] = ul + x + y;
	
	form_plane(camera, e[1], e[0], 0);
	form_plane(camera, e[0], e[2], 1);
	form_plane(camera, e[3], e[1], 2);
	form_plane(camera, e[2], e[3], 3);
	
	
}

void FrustrumClipper::form_plane(const Vector &a, const Vector &b, const Vector &c, int index)
{
	Vector x = b - a;
	Vector y = c - a;
	Vector nor = x ^ y;
	nor.norm();
	n[index] = nor;
	D[index] = -(a*nor);
}

double FrustrumClipper::eval(const Vector &v, int which)
{
	if (which == -1) which = defeval;
	return D[which] + (v * n[which]);
}

void FrustrumClipper::clip(ClipRes& cr, Vector trio[], int which)
{
	if (which == 0) {
		cr.n = 3;
		for (int i = 0; i < 3; i++) {
			cr.v[i] = trio[i];
			cr.bary[i] = imatrix[i];
		}
		return;
	}
	if (which & (which - 1)) {
		printf("%s: Clipping against more than one plane is (still) not supported\n", __FUNCTION__);
		cr.n = 0;
		return;
	}
	defeval = 0;
	while (0 == (which & (1 << defeval))) defeval++;
	
	cr.n = 0;
	bool vis[3];
	for (int i = 0; i < 3; i++) {
		vis[i] = eval(trio[i]) > -1;
	}
	for (int i = 0; i < 3; i++) {
		int j = (i + 1) % 3;
		if (vis[i]) {
			cr.v[cr.n] = trio[i];
			cr.bary[cr.n++] = imatrix[i];
		}
		if (vis[i] ^ vis[j]) {
			double av = eval(trio[i]);
			double bv = eval(trio[j]);
			double p = - av / (bv - av);
			if (p < 0) p = 0;
			if (p > 1) p = 1;
			cr.v[cr.n] = trio[i] * (1.0-p) + trio[j] * p;
			cr.bary[cr.n++] = imatrix[i] * (1.0-p) + imatrix[j] * p;
		}
	}
}


struct VistaBufferRasterizer {
	unsigned *b, ccolor;
	int xr, yr;
	Object **o;
	int n;
	FrustrumClipper * clipper;
	
	void doit(void)
	{
		for (int i = 0; i < n; i++) {
			draw_bbox(o[i]->get_bbox(), 1 << i);
		}
	}
	
	void draw_bbox(BBox box, unsigned color)
	{
		ccolor = color;
		Vector v[8];
		v[0] = box.minv;
		v[7] = box.maxv;
		for (int i = 1; i < 7; i++) {
			v[i] = v[0];
			for (int j = 0; j < 3; j++)
				if (i & (1 << j))
					v[i][j] = v[7][j];
		}
		draw_quad(v[0], v[1], v[3], v[2]);
		draw_quad(v[0], v[4], v[6], v[2]);
		draw_quad(v[0], v[4], v[5], v[1]);
		draw_quad(v[5], v[4], v[6], v[7]);
		draw_quad(v[5], v[1], v[3], v[7]);
		draw_quad(v[7], v[6], v[2], v[3]);
	}
	
	void draw_quad(const Vector & a, const Vector &b, const Vector &c, const Vector &d)
	{
		draw_trio(a, b, c);
		draw_trio(a, c, d);
	}
	
	void draw_trio(const Vector &a, const Vector &b, const Vector &c)
	{
		if (a == b || a == c || b == c) return;
		Vector l[16][3] = {{a, b, c}};
		int n = 1;
		for (int i = 0; i < 4; i++) {
			Vector q[16][3];
			int m = 0;
			int mask = 1 << i;
			for (int j = 0; j < n; j++) {
				Vector v[3] = {l[j][0], l[j][1], l[j][2]};
				ClipRes res;
				clipper->clip(res, v, mask);
				if (res.n) {
					q[m][0] = res.v[0];
					q[m][1] = res.v[1];
					q[m][2] = res.v[2];
					m++;
					if (res.n == 4) {
						q[m][0] = res.v[0];
						q[m][1] = res.v[2];
						q[m][2] = res.v[3];
						m++;
					}
				}
			}
			n = m;
			for (int j = 0; j < n; j++)
				for (int k = 0; k < 3; k++)
					l[j][k] = q[j][k];
		}
		for (int i = 0; i < n; i++) {
			draw_trio_really(l[i][0], l[i][1], l[i][2]);
		}
	}
	
	float lin_eval(const vec2f& a, const vec2f &b, float x)
	{
		if (a.y == b.y) return max(a.x, b.x);
		x = (x - a.y) / (b.y - a.y);
		return a.x + x * (b.x - a.x);
	}
	
	void draw_trio_really(const Vector &A, const Vector &B, const Vector &C)
	{
		vec2f a[3];
		a[0] = camera.reverse(A);
		a[1] = camera.reverse(B);
		a[2] = camera.reverse(C);
		//..
		#define SWCK(a, b) if (a.y > b.y) { vec2f t = a; a = b; b = t; }
		SWCK(a[0], a[1]);
		SWCK(a[1], a[2]);
		SWCK(a[0], a[1]);
		//..
		for (int y = (int) ceil(a[0].y); y <= (int) floor(a[2].y); y++) {
			if (y < 0 || y >= yr) continue;
			float x1, x2;
			x1 = lin_eval(a[0], a[2], y);
			if (y < a[1].y)
				x2 = lin_eval(a[0], a[1], y);
			else
				x2 = lin_eval(a[1], a[2], y);
			int ix0 = (int) floor(min(x1, x2));
			int ix1 = (int) ceil(max(x1, x2));
			unsigned *buff = b + (y * config.xres);
			for (int x = ix0; x <= ix1; x++) {
				if (x < 0 || x >= xr) continue;
				buff[x] |= ccolor;
			}
		}
	}
};


void draw_vista(unsigned *b, Object **o, int n)
{
	Vector ul, xdir, ydir, cam;
	Ray r1 = camera.get_screen_ray(0, 0);
	Ray r2 = camera.get_screen_ray(config.xres, 0);
	Ray r3 = camera.get_screen_ray(0, config.yres);
	cam = r1.c;
	ul = r1.c + r1.dir;
	xdir = r2.c + r2.dir - ul;
	ydir = r3.c + r3.dir - ul;
	FrustrumClipper clipper(cam, ul, xdir, ydir);
	VistaBufferRasterizer rasterizer;
	rasterizer.b = b;
	rasterizer.xr = config.xres;
	rasterizer.yr = config.yres;
	rasterizer.n = n;
	rasterizer.o = o;
	rasterizer.clipper = &clipper;
	rasterizer.doit();
}
